//
//  InverseKinematicsSolver.h
//  SimpleProg
//
//  Created by kumada on 2013/08/30.
//  Copyright (c) 2013年 kumada. All rights reserved.
//

#ifndef __SimpleProg__InverseKinematicsSolver__
#define __SimpleProg__InverseKinematicsSolver__

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <array>

class InverseKinematicsSolver
{
public:
        struct RotationMatrixAroundXAxis
        {
                const boost::numeric::ublas::matrix<double> operator()(double theta) const
                {
                        namespace ublas = boost::numeric::ublas;
                        ublas::matrix<double> tmp = identity_;
                        const double cost = std::cos(theta);
                        const double sint = std::sin(theta);
                        tmp(1, 1) = cost, tmp(1, 2) = -sint;
                        tmp(2, 1) = sint, tmp(2, 2) =  cost;
                        return tmp; // RVO of move
                }
        };
        
        struct RotationMatrixAroundYAxis
        {
                const boost::numeric::ublas::matrix<double> operator()(double theta) const
                {
                        namespace ublas = boost::numeric::ublas;
                        ublas::matrix<double> tmp = identity_;
                        const double cost = std::cos(theta);
                        const double sint = std::sin(theta);
                        tmp(0, 0) = cost,  tmp(0, 2) = sint;
                        tmp(2, 0) = -sint, tmp(2, 2) = cost;
                        return tmp;
                }
        };

        struct RotationMatrixAroundZAxis
        {
                const boost::numeric::ublas::matrix<double> operator()(double theta) const
                {
                        namespace ublas = boost::numeric::ublas;
                        ublas::matrix<double> tmp = identity_;
                        const double cost = std::cos(theta);
                        const double sint = std::sin(theta);
                        tmp(0, 0) = cost, tmp(0, 1) = -sint;
                        tmp(1, 0) = sint, tmp(1, 1) =  cost;
                        return tmp;
                }
        };

        struct DifferentialRotationMatrixAroundXAxis
        {
                const boost::numeric::ublas::matrix<double> operator()(double theta) const
                {
                        namespace ublas = boost::numeric::ublas;
                        ublas::matrix<double> tmp = zero_;
                        const double cost = std::cos(theta);
                        const double sint = std::sin(theta);
                        tmp(1, 1) = -sint, tmp(1, 2) = -cost;
                        tmp(2, 1) =  cost, tmp(2, 2) = -sint;
                        return tmp; // RVO of move
                }
        };

        struct DifferentialRotationMatrixAroundYAxis
        {
                const boost::numeric::ublas::matrix<double> operator()(double theta) const
                {
                        namespace ublas = boost::numeric::ublas;
                        ublas::matrix<double> tmp = zero_;
                        const double cost = std::cos(theta);
                        const double sint = std::sin(theta);
                        tmp(0, 0) = -sint, tmp(0, 2) =  cost;
                        tmp(2, 0) = -cost, tmp(2, 2) = -sint;
                        return tmp;
                }
        };

        struct DifferentialRotationMatrixAroundZAxis
        {
                const boost::numeric::ublas::matrix<double> operator()(double theta) const
                {
                        namespace ublas = boost::numeric::ublas;
                        ublas::matrix<double> tmp = zero_;
                        const double cost = std::cos(theta);
                        const double sint = std::sin(theta);
                        tmp(0, 0) = -sint, tmp(0, 1) = -cost;
                        tmp(1, 0) =  cost, tmp(1, 1) = -sint;
                        return tmp;
                }
        };
        
        struct TranslationMatrix
        {
                const boost::numeric::ublas::matrix<double> operator()(const boost::numeric::ublas::vector<double>& v) const
                {
                        namespace ublas = boost::numeric::ublas;
                        ublas::matrix<double> tmp = identity_;
                        tmp(0, 3) = v(0);
                        tmp(1, 3) = v(1);
                        tmp(2, 3) = v(2);
                        return tmp;
                }
        };

        const TranslationMatrix translation_matrix_ {};
        
        const std::tuple<
                RotationMatrixAroundXAxis,
                RotationMatrixAroundYAxis,
                RotationMatrixAroundZAxis
        > rotational_matrices_
        {
                std::make_tuple(
                        RotationMatrixAroundXAxis{},
                        RotationMatrixAroundYAxis{},
                        RotationMatrixAroundZAxis{}
                )
        };
                
        const std::tuple<
                DifferentialRotationMatrixAroundXAxis,
                DifferentialRotationMatrixAroundYAxis,
                DifferentialRotationMatrixAroundZAxis
        > differential_rotational_matrices_
        {
                std::make_tuple(
                        DifferentialRotationMatrixAroundXAxis{},
                        DifferentialRotationMatrixAroundYAxis{},
                        DifferentialRotationMatrixAroundZAxis{}
                )
        };
        
        std::array<boost::numeric::ublas::vector<double>, 3> axises_;
        
        template<int R, int T>
        struct Arm
        {
                static constexpr int Rotation_ = R;
                static constexpr int Translation_ = T;
                double angle_;
                double length_;
                
                Arm()
                        : angle_(0)
                        , length_(0) {}
                
        };
        
        enum RotationDirection
        {
                RX,
                RY,
                RZ,
        };

        enum TranslationDirection
        {
                TX,
                TY,
                TZ,
        };
        
        enum Axis
        {
                X,
                Y,
                Z
        };
        
        
        struct Arms
        {
                Arm<RZ, TZ> s_;
                Arm<RY, TZ> l_;
                Arm<RY, TZ> u_;
                Arm<RZ, TZ> r_;
                Arm<RY, TZ> b_;
                Arm<RZ, TZ> t_;
        };
        
        InverseKinematicsSolver();
        ~InverseKinematicsSolver() = default;
        
        void execute(const boost::numeric::ublas::vector<double>& dp, Arms& arms) const;
        
private:
        InverseKinematicsSolver(const InverseKinematicsSolver&) = delete;
        InverseKinematicsSolver& operator=(const InverseKinematicsSolver&) = delete;
        InverseKinematicsSolver(InverseKinematicsSolver&&) = delete;
        InverseKinematicsSolver operator=(InverseKinematicsSolver&&) = delete;

        
        boost::numeric::ublas::vector<double> one_;
        static const boost::numeric::ublas::identity_matrix<double>     identity_;
        static const boost::numeric::ublas::zero_matrix<double>         zero_;
        
        template<int R, int T>
        const boost::numeric::ublas::matrix<double> prod(Arm<R, T>& arm) const
        {
                namespace ublas = boost::numeric::ublas;
                return ublas::prod(
                        std::get<R>(rotational_matrices_)(arm.angle_),
                        translation_matrix_(arm.length_ * axises_[T])
                );
        }

        template<int R, int T>
        const boost::numeric::ublas::matrix<double> diff_prod(Arm<R, T>& arm) const
        {
                namespace ublas = boost::numeric::ublas;
                return ublas::prod(
                        std::get<R>(differential_rotational_matrices_)(arm.angle_),
                        translation_matrix_(arm.length_ * axises_[T])
                );
        }
        
        
        const boost::numeric::ublas::vector<double> six_prod(
                const boost::numeric::ublas::matrix<double>& m0,
                const boost::numeric::ublas::matrix<double>& m1,
                const boost::numeric::ublas::matrix<double>& m2,
                const boost::numeric::ublas::matrix<double>& m3,
                const boost::numeric::ublas::matrix<double>& m4,
                const boost::numeric::ublas::matrix<double>& m5
        ) const;

};
#endif /* defined(__SimpleProg__InverseKinematicsSolver__) */
