//
//  RobotData.cpp
//  SimpleProg
//
//  Created by kumada on 2013/08/21.
//  Copyright (c) 2013年 kumada. All rights reserved.
//

#include "RobotData.h"
#include <string>
#include <list>
#include <fstream>
#include <iostream>
#include <boost/algorithm/string.hpp>
#include <boost/range.hpp>
#include <array>

#ifdef __CYGWIN__
#include <boost/lexical_cast.hpp>
#define LEXICAL_CAST_INT(x)     boost::lexical_cast<int>(x)
#define LEXICAL_CAST_DOUBLE(x)  boost::lexical_cast<double>(x)
#endif

#ifdef __APPLE__
#define LEXICAL_CAST_INT(x)     std::stoi(x)
#define LEXICAL_CAST_DOUBLE(x)  std::stof(x)
#endif

void RobotData::load(const std::string& filename)
{
        std::ifstream ifs {filename};
        std::string line;
        
        // read the number of bodies
        std::getline(ifs, line);
        boost::trim(line);
        int body_size = LEXICAL_CAST_INT(line);
        bodies_.resize(body_size);
        
        for ( int i = 0; i < body_size; ++i ) {
        
                bodies_[i] = BodyPtr{ new Body {} };
                
                // skip one line
                std::getline(ifs, line);
                
                Body* body = bodies_[i].get();
                
                // read faces belonging to the body
                read_faces(body, ifs, line);
                
                // read edges belonging to the body
                read_edges(body, ifs, line);
                
                // read vertices belonging to the body
                read_vertices(body, ifs, line);
        }
}

const std::vector<int> RobotData::extract_integers(const std::string& line) const
{
        auto begin = REGEX_NAMESPACE::sregex_iterator {line.begin(), line.end(), int_regex_};
        auto end = REGEX_NAMESPACE::sregex_iterator {};
        auto range = boost::make_iterator_range(begin, end);
        std::vector<int> numbers;
        numbers.reserve(std::distance(begin, end));
        for ( auto match : range ) {
                std::string match_str = match.str();
                int number = LEXICAL_CAST_INT(match_str);
                numbers.push_back(number);
        }
        return numbers;
}

const std::vector<double> RobotData::extract_reals(const std::string& line) const
{
        auto begin = REGEX_NAMESPACE::sregex_iterator {line.begin(), line.end(), real_regex_};
        auto end = REGEX_NAMESPACE::sregex_iterator {};
        auto range = boost::make_iterator_range(begin, end);
        std::vector<double> numbers;
        numbers.reserve(std::distance(begin, end));
        for ( auto match : range ) {
                std::string match_str = match.str();
                double real = LEXICAL_CAST_DOUBLE(match_str);
                numbers.push_back(real);
        }
        return numbers;
}

void RobotData::read_faces(Body* body, std::ifstream& ifs, std::string& line)
{
        std::getline(ifs, line);
        boost::trim(line);
        int face_size = LEXICAL_CAST_INT(line);
        std::vector<Body::FacePtr>& faces = body->get_faces();
        faces.resize(face_size);
        for ( int i = 0; i < face_size; ++i ) {
                faces[i] = Body::FacePtr{ new Face {} };
                Face* face = faces[i].get();

                std::getline(ifs, line);
                const std::vector<int> n = extract_integers(line);
                face->set_edge(n[0]);
                face->set_next(n[1]);
        }
}

void RobotData::read_edges(Body* body, std::ifstream& ifs, std::string& line)
{
        std::getline(ifs, line);
        boost::trim(line);
        int edge_size = LEXICAL_CAST_INT(line);
        std::vector<Body::EdgePtr>& edges = body->get_edges();
        edges.resize(edge_size);
        for ( int i = 0; i < edge_size; ++i ) {
                edges[i] = Body::EdgePtr{ new Edge {} };
                Edge* edge = edges[i].get();
                
                std::getline(ifs, line);
                const std::vector<int> n0 = extract_integers(line);
                edge->set_start_vertex(n0[0]);
                edge->set_end_vertex(n0[1]);
                
                std::getline(ifs, line);
                const std::vector<int> n1 = extract_integers(line);
                edge->set_start_clockwise_edge(n1[0]);
                edge->set_end_clockwise_edge(n1[1]);
                edge->set_right_face(n1[2]);
                edge->set_left_face(n1[3]);
        }
}

void RobotData::read_vertices(Body* body, std::ifstream& ifs, std::string& line)
{
        std::getline(ifs, line);
        boost::trim(line);
        int vertex_size = LEXICAL_CAST_INT(line);
        std::vector<Body::VertexPtr>& vertices = body->get_vertices();
        vertices.resize(vertex_size);
        for ( int i = 0; i < vertex_size; ++i ) {
                std::getline(ifs, line);
                const std::vector<double> rs = extract_reals(line);
                vertices[i] = Body::VertexPtr{ new Vertex {rs[0], rs[1], rs[2]} };
        }
}
