//
//  robot_simulator.cpp
//  SimpleProg
//
//  Created by kumada on 2013/08/21.
//  Copyright (c) 2013年 kumada. All rights reserved.
//

#include "robot_simulator.h"
#include "RobotData.h"
#include "MatrixStack.h"
#include "SimulatorProperties.h"
#include "GeometryDescriptor.h"
#include "Window.h"
#include "InverseKinematicsSolver.h"
#include <boost/algorithm/string.hpp>
#include <boost/range/algorithm/for_each.hpp>
#include <boost/math/constants/constants.hpp>
#include <array>
#include <map>

#ifdef __APPLE__
#include <OpenGL/OpenGL.h>
#include <GLUT/GLUT.h>
#endif

#ifdef __CYGWIN__
#include <windows.h>
#include <gl/gl.h>
#include <gl/glut.h>
#endif

#define MOVE

namespace ublas = boost::numeric::ublas;

namespace
{
        RobotData robot_data_ {};
        InverseKinematicsSolver::Arms arms_;

        SimulatorProperties simulator_properties_ {arms_};
        InverseKinematicsSolver solver_ {};
        
        std::array<ublas::vector<double>, 8> targets_;
        ublas::vector<double> current_(3);
        int target_index_ = 0;

        int step_ = 1;
        
        constexpr double Alpha = 0.1;
        constexpr double Epsilon = 0.1;
        
        double u_length_;
        double r_length_;
        double b_length_;
        double t_length_;
        
        // Light position
        constexpr float Light0Position[] = {0.0f, 0.0f, 1.0f, 0.0f}; // light from infinite distance
        constexpr float Light1Position[] = {100.0f, 100.0f, 0.0f, 1.0f}; // point light
        
        // diffuse component
        constexpr float Light0Diffuse[] = {0.8, 0.8, 0.8, 1.0};
        constexpr float Light1Diffuse[] = {0.5, 0.5, 0.5, 1.0};
        
        // specular component
        constexpr float LightSpecular[] = {0.3, 0.3, 0.3, 1.0};
        
        // ambient light
        constexpr float LightModelAmbient[]  = {0.1, 0.1, 0.1, 0.1};
        
        // pi
        constexpr double Pi = boost::math::constants::pi<double>();
        
        typedef std::map<GLenum, const float*> LightProperties;
        
        void create_light(GLenum name, const LightProperties& properties)
        {
                for ( const auto& property : properties ) {
                        glLightfv(name, property.first, property.second);
                }
        }

        void initialize()
        {
                // light0
                const LightProperties light0 = {
                        {GL_POSITION, Light0Position},
                        {GL_DIFFUSE,  Light0Diffuse },
                        {GL_SPECULAR, LightSpecular }
                };
                create_light(GL_LIGHT0, light0);
                
                // light 1
                const LightProperties light1 = {
                        {GL_POSITION, Light1Position},
                        {GL_DIFFUSE,  Light1Diffuse },
                        {GL_SPECULAR, LightSpecular }
                };
                create_light(GL_LIGHT1, light1);
                
                // environment light
                glLightModelfv(GL_LIGHT_MODEL_AMBIENT, LightModelAmbient);
                
                glEnable(GL_LIGHTING);
                glEnable(GL_LIGHT0);
                glEnable(GL_LIGHT1);
                
                glDepthFunc(GL_LEQUAL);
                glEnable(GL_DEPTH_TEST);

        }
        
        constexpr double Eye[] = {200.0, 200.0, 100.0};
        constexpr double Center[] = {0.0, 0.0, 0.0};

        void calculate_normal(const double v[], double up[])
        {
                up[0] = -v[0] * v[2];
                up[1] = -v[1] * v[2];
                up[2] = v[0] * v[0] + v[1] * v[1];
        }
        
        void make_unit_vector(double up[])
        {
                double norm = up[0] * up[0] + up[1] * up[1] + up[2] * up[2];
                double length = std::sqrt(norm);
                for ( int i = 0; i < 3; ++i ) {
                        up[i] /= length;
                }
        }

        void make_viewup(double up[])
        {
                double vector[3];
                for ( int i = 0; i < 3; ++i ) {
                        vector[i] = Center[i] - Eye[i];
                }
                
                calculate_normal(vector, up);
                make_unit_vector(up);
        }

        void reshape(int w, int h)
        {
                glViewport(0, 0, w, h);
                
                glMatrixMode(GL_PROJECTION);
                glLoadIdentity();
                
                gluPerspective(30.0, w / (static_cast<double>(h)), 10.0, 1000.0);
                
                glMatrixMode(GL_MODELVIEW);
                glLoadIdentity();

                double up[3];
                make_viewup(up);
                gluLookAt(Eye[0], Eye[1], Eye[2], Center[0], Center[1], Center[2], up[0], up[1], up[2]);

        }
        
        void set_initial_state();
        
        void keyboard(std::uint8_t key, int x, int y)
        {
                switch (key) {
                case 27:
                        std::exit(0);
                        break;
                case 'a':
                        simulator_properties_.add_s();
                        break;
                case 'b':
                        simulator_properties_.subtract_s();
                        break;
                case 'c':
                        simulator_properties_.add_l();
                        break;
                case 'd':
                        simulator_properties_.subtract_l();
                        break;
                case 'e':
                        simulator_properties_.add_u();
                        break;
                case 'f':
                        simulator_properties_.subtract_u();
                        break;
                case 'g':
                        simulator_properties_.add_r();
                        break;
                case 'h':
                        simulator_properties_.subtract_r();
                        break;
                case 'i':
                        simulator_properties_.add_b();
                        break;
                case 'j':
                        simulator_properties_.subtract_b();
                        break;
                case 'k':
                        simulator_properties_.add_t();
                        break;
                case 'l':
                        simulator_properties_.subtract_t();
                        break;

                default:
                        break;
                }
                glutPostRedisplay();
        }

        void keyboard2(std::uint8_t key, int x, int y)
        {
                switch (key) {
                case 27:
                        std::exit(0);
                        break;
                case 'a':
                        step_ = 1;
                        break;
                case 'b':
                        step_ = -1;
                        break;
                case 'c':
                        glutIdleFunc(glutPostRedisplay);
                        break;
                case 'd':
                        glutIdleFunc(0);
                        break;

                default:
                        break;
                }
                glutPostRedisplay();
        }
        
        
        constexpr float MaterialColorRed[]      = {1.0f, 0.0f, 0.0f, 1.0f};
        constexpr float MaterialColorBlue[]     = {0.0f, 0.0f, 1.0f, 1.0f};
        constexpr float MaterialColorGreen[]    = {0.0f, 1.0f, 0.0f, 1.0f};
        constexpr float MaterialColorWhite[]    = {1.0f, 1.0f, 1.0f, 1.0f};
        
        
        constexpr float MaterialSpecular[]      = {0.2f, 0.2f, 0.2f, 1.0f};

        double to_degree(double radian)
        {
                return radian / boost::math::constants::pi<double>() * 180;
        }

        const Vertex::Point calculate_normal(const std::vector<const Vertex*>& vertices)
        {
                using namespace geo;
                const Vertex* v0 = vertices[0];
                const Vertex* v1 = vertices[1];
                const Vertex* v2 = vertices[2];
                
                typedef Vertex::Point Point;
                const Point& p0 = v0->get_point();
                const Point& p1 = v1->get_point();
                const Point& p2 = v2->get_point();
                
                const Point p10 = p0 - p1;
                const Point p12 = p2 - p1;
                return geo::unit<double, ThreeDimension>(geo::outer_product(p12, p10));
        }
        
        void draw_face(const Body* body, int i)
        {
                typedef Vertex::Point Point;
                const std::vector<const Vertex*> vertices = body->extract_vertices(i);
                const Point normal = calculate_normal(vertices);
                
                {
                        GeometryDescriptor descriptor{GL_POLYGON};
                        glNormal3dv(normal.data());
                        
                        for ( auto vertex : vertices ) {
                                const auto& p = vertex->get_point();
                                glVertex3dv(p.data());
                        }
                }
                
                
        }
        
        void draw_arm_with_shading(const Body* body)
        {
                std::size_t face_size = body->get_face_size();
                for ( int i = 0; i < static_cast<int>(face_size); ++i ) {
                        draw_face(body, i);
                }
        }
        
        template<int R, int T>
        void rotate(const InverseKinematicsSolver::Arm<R, T>& arm)
        {
                const auto& axis = solver_.axises_[arm.Rotation_];
                glRotatef(to_degree(arm.angle_), axis(0), axis(1), axis(2));
        }

        void draw_s(const Body* s_arm)
        {
                rotate(arms_.s_);
                
                // red
                glMaterialfv(GL_FRONT, GL_DIFFUSE, MaterialColorRed);
                glMaterialf(GL_FRONT, GL_SHININESS, 10);
                draw_arm_with_shading(s_arm);
        }

        void draw_l(const Body* l_arm)
        {
                glTranslatef(0, 0, arms_.s_.length_);
                rotate(arms_.l_);
                glTranslatef(0, 0, -arms_.s_.length_);
                
                // blue
                glMaterialfv(GL_FRONT, GL_DIFFUSE, MaterialColorBlue);
                draw_arm_with_shading(l_arm);
        }
        
        void draw_u(const Body* u_arm)
        {
                glTranslatef(0, 0, u_length_);
                rotate(arms_.u_);
                glTranslatef(0, 0, -u_length_);
                
                // blue
                glMaterialfv(GL_FRONT, GL_DIFFUSE, MaterialColorGreen);
                draw_arm_with_shading(u_arm);
        }
        
        void draw_r(const Body* r_arm)
        {
                glTranslatef(0, 0, r_length_);
                rotate(arms_.r_);
                glTranslated(0, 0, -r_length_);
                
                glMaterialfv(GL_FRONT, GL_DIFFUSE, MaterialColorBlue);
                draw_arm_with_shading(r_arm);
                
        }

        void draw_b(const Body* b_arm)
        {
                glTranslatef(0, 0, b_length_);
                rotate(arms_.b_);
                glTranslated(0, 0, -b_length_);
                
                glMaterialfv(GL_FRONT, GL_DIFFUSE, MaterialColorGreen);
                draw_arm_with_shading(b_arm);
                
        }
        
        void draw_t(const Body* t_arm)
        {
                glTranslatef(0, 0, t_length_);
                rotate(arms_.t_);
                glTranslated(0, 0, -t_length_);
                
                glMaterialfv(GL_FRONT, GL_DIFFUSE, MaterialColorRed);
                draw_arm_with_shading(t_arm);
        }
        
        
        
        void cylinger_along_y_axis(double radius, double length, int num, const float color[])
        {
                MatrixStack stack {};
                glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
                glMaterialf(GL_FRONT, GL_SHININESS, 10);
                {
                        GeometryDescriptor descriptor {GL_QUAD_STRIP};
                        
                        double step = 2 * Pi / num;
                        for ( int i = 0; i <= num; ++i ) {
                                const double t = step * i;
                                const double cost = std::cos(t);
                                const double sint = std::sin(t);
                                glNormal3d(cost, 0, sint);
                                const double rcost = radius * cost;
                                const double rsint = radius * sint;
                                glVertex3d(rcost, -length, rsint);
                                glVertex3d(rcost,  length, rsint);
                        }
                }
                glTranslated(0, length, 0);
                glRotated(-90, 1, 0, 0);
                glutSolidCone(2 * radius, 4 * radius, 5, 5);
        }
        
        void cylinger_along_z_axis(double radius, double length, int num, const float color[])
        {
                MatrixStack stack {};
                glRotated(90, 1, 0, 0);
                cylinger_along_y_axis(radius, length, num, color);
        }

        void cylinger_along_x_axis(double radius, double length, int num, const float color[])
        {
                MatrixStack stack {};
                glRotated(-90, 0, 0, 1);
                cylinger_along_y_axis(radius, length, num, color);
        }
        
        void origin(double radius, double slices, double stacks, const float color[])
        {
                glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
                glMaterialf(GL_FRONT, GL_SHININESS, 10);
                glutSolidSphere(radius, slices, stacks);
        }
        
        void draw_coordinate_axis(double length)
        {
                cylinger_along_x_axis(1, length, 6, MaterialColorRed);
                cylinger_along_y_axis(1, length, 6, MaterialColorGreen);
                cylinger_along_z_axis(1, length, 6, MaterialColorBlue);
                origin(2, 6, 6, MaterialColorWhite);
        }
        
        void draw_target(double radius, const float color[])
        {
                MatrixStack stack{};
                glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
                glMaterialf(GL_FRONT, GL_SHININESS, 10);

                glTranslated(targets_[target_index_](0), targets_[target_index_](1), targets_[target_index_](2));
                glutSolidSphere(radius, 16, 16);
        }

        void draw_current(double radius, const float color[])
        {
                MatrixStack stack{};
                glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
                glMaterialf(GL_FRONT, GL_SHININESS, 10);

                glTranslated(current_(0), current_(1), current_(2));
                glutSolidSphere(radius, 16, 16);
        }
        
        void move_arms()
        {
                ublas::vector<double> p = targets_[target_index_] - current_;
                double norm_p = ublas::norm_2(p);
                if ( norm_p < Epsilon ) {
                
                        current_ = targets_[target_index_];
                        //target_index_ = get_next_target_index(target_index_);//(target_index_ + 1) % targets_.size();
                        target_index_ = (target_index_ + step_) % targets_.size();
                        
                        p = targets_[target_index_] - current_;
                        norm_p = ublas::norm_2(p);
                }
                ublas::vector<double> dp = Alpha * (p / norm_p);
                solver_.execute(dp, arms_);
                current_ += dp;
        }
        
        void display()
        {
                // clear buffer
                glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
                
                // set specular
                glMaterialfv(GL_FRONT, GL_SPECULAR, MaterialSpecular);
                
                draw_coordinate_axis(70);
                draw_target(3, MaterialColorGreen);
                //draw_current(3, MaterialColorBlue);
                {
                        MatrixStack stack0 {};
                        
                        
                        draw_s(robot_data_.get_body(0));
                        draw_l(robot_data_.get_body(1));
                        draw_u(robot_data_.get_body(2));
                        draw_r(robot_data_.get_body(3));
                        draw_b(robot_data_.get_body(4));
                        draw_t(robot_data_.get_body(5));
                }

                glFlush();
                glutSwapBuffers();
#ifdef MOVE
                try {
                        move_arms();
                } catch (const std::exception& ex) {
                        std::cout << ex.what();
                }
#endif
        }
        
        void set_initial_state()
        {
                arms_.s_.angle_ = 0.0;
                arms_.l_.angle_ = 0.0;
                arms_.u_.angle_ = ::Pi / 2.0;
                arms_.r_.angle_ = 0.0;
                arms_.b_.angle_ = 0.0;
                arms_.t_.angle_ = 0.0;
                
                current_(0) = 57.5, current_(1) = 0, current_(2) = 37.5;
        }
}

int robot_simulator(int argc, char * argv[])
{
        const std::string filename = argv[1];
        robot_data_.load(filename);

        for ( auto& target : targets_ ) {
                target.resize(3);
        }
        
        targets_[0](0) = 0,   targets_[0](1) = 50,  targets_[0](2) = 30;
        targets_[1](0) = -25, targets_[1](1) = 25,  targets_[1](2) = 50;
        targets_[2](0) = -50, targets_[2](1) = 0,   targets_[2](2) = 30;
        targets_[3](0) = -25, targets_[3](1) = -25, targets_[3](2) = 50;
        targets_[4](0) = 0,   targets_[4](1) = -50, targets_[4](2) = 30;
        targets_[5](0) = 25,  targets_[5](1) = -25, targets_[5](2) = 50;
        targets_[6](0) = 50,  targets_[6](1) = 0,   targets_[6](2) = 30;
        targets_[7](0) = 25,  targets_[7](1) = 25,  targets_[7](2) = 50;

        arms_.s_.length_ = 17.5;
        arms_.l_.length_ = 20;
        arms_.u_.length_ = 15;
        arms_.r_.length_ = 15;
        arms_.b_.length_ = 15;
        arms_.t_.length_ = 12.5;

        u_length_ = arms_.s_.length_ + arms_.l_.length_;
        r_length_ = u_length_ + arms_.u_.length_;
        b_length_ = r_length_ + arms_.r_.length_;
        t_length_ = b_length_ + arms_.b_.length_;
        
        set_initial_state();
        
        glutInit(&argc, argv);
        glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
        glutInitWindowSize(Window::Width_, Window::Height_);
        glutInitWindowPosition(Window::Left_, Window::Top_);
        glutCreateWindow("Robot Model");
        initialize();
        glutReshapeFunc(reshape);
#ifdef MOVE
        glutKeyboardFunc(keyboard2);
#else
        glutKeyboardFunc(keyboard);
#endif
        glutDisplayFunc(display);
        
        glutMainLoop();
        return 0;
}
